/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: osNetBSD14.c,v 7.0.2.1 2000/11/29 15:04:35 cvs Exp $
 *
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Defines
 */

/*
 * Globals:
 */

extern int iti_version_code;	/* kept in itifgMain.c for consistency */
extern const struct mod_names_t mod_names[4]; /* itifgAux.c */

static int
iti_os_probe __P((struct device*, struct cfdata*, void*));
static void
iti_os_attach __P((struct device*, struct device*, void*));

struct cfattach bktr_ca = {
        sizeof(itifg_t), iti_os_probe, iti_os_attach
};

extern struct cfdriver bktr_cd;

#if (0)
pgprot_t protection_map[16] = {
	__P000, __P001, __P010, __P011, __P100, __P101, __P110, __P111,
	__S000, __S001, __S010, __S011, __S100, __S101, __S110, __S111
};
#endif

/*
 * Prototypes:
 */

int
iti_os_interrupt (void*);
static void
iti_os_cam_timeout (void*);
static void
iti_os_iop_timeout (void*);
#if (0)
static void
iti_os_task_queue (void*);
#endif

#define iti_os_open	itifgopen
#define iti_os_close	itifgclose
#define iti_os_read	itifgread
#define iti_os_write	itifgwrite
#define iti_os_ioctl	itifgioctl
#define iti_os_poll	itifgpoll
#define iti_os_mmap	itifgmmap

#define iti_os_llseek	itifgllseek
#define iti_os_fasyc	itifgfasync

int
iti_os_open (dev_t, int, int, struct proc*);
int
iti_os_close (dev_t, int, int, struct proc*);
int
iti_os_read (dev_t, struct uio*, int);
int
iti_os_write (dev_t, struct uio*, int);
int
iti_os_ioctl (dev_t, u_long, caddr_t, int, struct proc*); 
int
iti_os_poll (dev_t, int, struct proc*);
paddr_t
iti_os_mmap (dev_t, off_t, int);

/* FIXME */
#if (0)
off_t
iti_os_llseek (struct inode*, off_t, int);
int
iti_os_fasync (struct inode*, struct file*, int);
#endif

/*
 * iti_malloc
 */

void* 
iti_malloc (size_t size, iti_alloc_t mode)
{
  switch (mode)
    {
    case ITI_UNPAGED:
      return malloc (size, M_DEVBUF, 0);
      break;
    case ITI_PAGED:
      /* return vmalloc (size); */ /* FIXME */
      return NULL;
      break;
    default:
      return NULL;
    }
  
  return NULL;
}

/*
 * iti_free
 */

void
iti_free(void *base, iti_alloc_t mode)
{
  switch (mode)
    {
    case ITI_UNPAGED:
      return free (base, M_DEVBUF);
      break;
    case ITI_PAGED:
      /* return vfree (base); */ /* FIXME */
      break;
    }
}

/*
 * iti_add_timer
 */

int
iti_add_timer (itifg_t *iti, time_t when, iti_sleep_t mode)
{
  void (*func) (void*) = NULL;

  switch (mode)
    {
    case ITI_CAMERA:
      func = iti_os_cam_timeout;
      break;
    case ITI_IOPORT:
      func = iti_os_iop_timeout;
      break;
    default:
      return -EINVAL;
    }

  untimeout (func, iti);

  timeout (func, iti, time.tv_usec + when * 10000); /* FIXME */

  return OK;
}

/*
 * iti_del_timer
 */

int
iti_del_timer (itifg_t *iti, iti_sleep_t mode)
{
  void (*func) (void*) = NULL;

  switch (mode)
    {
    case ITI_CAMERA:
      func = iti_os_cam_timeout;
      break;
    case ITI_IOPORT:
      func = iti_os_iop_timeout;
      break;
    default:
      return -EINVAL;
    }
  
  untimeout (func, iti);
  
  return OK;
}

/*
 * iti_sleep
 */

int
iti_sleep (itifg_t *iti, time_t when, iti_sleep_t mode)
{
  int error;

  switch (mode)
    {
    case ITI_CAMERA:
      if (iti->osp.cam_w_wait)
	return -EBUSY;
      iti->osp.cam_w_wait = TRUE;
      error = tsleep (&iti->osp.cam_w_wait, ITIFGPRI, "frame",
		     time.tv_usec + when * 10000); /* FIXME */
      iti->osp.cam_w_wait = FALSE;
      break;
    case ITI_IOPORT:
      if (iti->osp.iop_w_wait)
	return -EBUSY;
      iti->osp.iop_w_wait = TRUE;
      error = tsleep (&iti->osp.iop_w_wait, ITIFGPRI, "ioport",
		      time.tv_usec + when * 10000); /* FIXME */
      iti->osp.iop_w_wait = FALSE;
      break;
    default:
      return -EINVAL;
    }
  
  if (error == ERESTART)
    return -ERESTART;
  
  return OK;
}

/*
 * iti_block_wakeup
 */

int
iti_block_wakeup (itifg_t *iti, iti_sleep_t mode)
{
  switch (mode)
    {
    case ITI_CAMERA:
      if (!iti->osp.cam_w_wait)
	return ERROR;
      wakeup (&iti->osp.cam_w_wait);
      break;
    case ITI_IOPORT:
      if (!iti->osp.cam_w_wait)
	return ERROR;
      wakeup (&iti->osp.iop_w_wait);
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * iti_select_wakeup
 */

int
iti_select_wakeup (itifg_t *iti, iti_sleep_t mode)
{

  switch (mode)
    {
    case ITI_CAMERA:
      if (!iti->osp.cam_s_wait)
	return ERROR;
      selwakeup (&iti->osp.cam_s_queue);
      break;
    case ITI_IOPORT:
      return -ENXIO; /* can't select on ioports yet */
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

#if (0) /* How to handle this? */
/*
 * iti_marktq
 */

int
iti_marktq (itifg_t *iti)
{
  iti->osp.clr_t_queue.sync = 0;
  iti->osp.clr_t_queue.routine = iti_os_task_queue;
  iti->osp.clr_t_queue.data = (void*)iti;
  
  queue_task (&iti->osp.clr_t_queue, &tq_immediate);
  mark_bh (IMMEDIATE_BH);  

  return OK;
}
#endif

/*
 * iti_lock
 */

void
iti_lock (itifg_t *iti)
{
  /* only for smp issues */
}

/*
 * iti_unlock
 */

void
iti_unlock (itifg_t *iti)
{
  /* only for smp issues */
}


/*
 * iti_irq_lock
 */

void
iti_irq_lock (itifg_t *iti)
{
  iti->osp.s = splbio ();
}

/*
 * iti_irq_unlock
 */

void
iti_irq_unlock (itifg_t *iti)
{
  splx (iti->osp.s);
}

/*
 * iti_get_time
 */

int
iti_get_time (struct timeval *when)
{
  if (!when)
    return ERROR;

  when->tv_sec = time.tv_sec;
  when->tv_usec = time.tv_usec;

  return OK;
}

/*
 * iti_get_major
 */

int
iti_get_major (iti_dev_t dev, int *major)
{
  if (!major)
    return ERROR;

  *major = major(dev);
  return OK;
}

/*
 * iti_get_minor
 */

int
iti_get_minor (iti_dev_t dev, int *minor)
{
  if (!minor)
    return ERROR;

  *minor = minor(dev);
  return OK;
}

#if (0) /* this should be handled by uvm */
/*
 * iti_virt_to_phys
 */

int
iti_virt_to_phys (caddr_t arg_virt,
		  u_long *ret_phys,
		  u_long *ret_bus,
		  int user_space)
{
  pgd_t	*pgd;
  pmd_t	*pmd;
  pte_t	*pte;

  u_long virt;
  void* kvirt;

  if (user_space)
    {
      virt = (u_long)arg_virt;
      pgd = pgd_offset (current->mm, virt);
    }
  else /* kernel_space */
    {
      virt = VMALLOC_VMADDR(arg_virt);
#if LINUX_VERSION_CODE < 0x020100
      pgd = pgd_offset (current->mm, virt);
#elif LINUX_VERSION_CODE >= 0x020200
      pgd = pgd_offset_k (virt);
#endif
    }

  if (pgd_none(*pgd)) return EFAULT;
  pmd = pmd_offset (pgd, virt);
  if (pmd_none(*pmd)) return EFAULT;
  pte = pte_offset (pmd, virt);
  if (!pte_present(*pte)) return EFAULT;

  kvirt = (void*)(pte_page(*pte) + (virt & ~PAGE_MASK));

  if (ret_phys)
    *ret_phys = virt_to_phys (kvirt);

  /* TESTME under 2.0 */
  if (ret_bus)
    *ret_bus = virt_to_bus (kvirt);
  
  return OK;
}

/*
 * iti_lock_page
 */

int
iti_lock_page (u_long page, int user_space)
{
  void *kvirt = phys_to_virt (page);

  if (!kvirt)
    return ERROR;


#if LINUX_VERSION_CODE < 0x020100
  iti_memset (kvirt, 0, PAGE_SIZE);
#elif LINUX_VERSION_CODE >= 0x020200
#  if defined(__powerpc__) || defined(__alpha__)
  clear_page ((long)kvirt);
#  else
  clear_page (kvirt);
#  endif
#endif

  if (user_space)
    set_bit(PG_locked, &mem_map[MAP_NR(kvirt)].flags);
  else /* kernel space */
    set_bit(PG_reserved, &mem_map[MAP_NR(kvirt)].flags);

  atomic_inc(&(mem_map[MAP_NR(kvirt)].count));

  return OK;
}

/*
 * iti_unlock_page
 */

int
iti_unlock_page (u_long page, int user_space)
{
  void *kvirt = phys_to_virt (page);

  if (!kvirt)
    return ERROR;

  if (user_space)
    clear_bit(PG_locked, &mem_map[MAP_NR(kvirt)].flags);    
  else /* kernel space */
    clear_bit(PG_reserved, &mem_map[MAP_NR(kvirt)].flags);

  atomic_dec(&(mem_map[MAP_NR(kvirt)].count));

  return OK;
}

/*
 * iti_remap_page
 */

int
iti_remap_page (caddr_t dst, u_long src)
{
  int error;
  pgprot_t pgprot;

  if (!dst || !src)
    return ERROR;
  
  pgprot_val(pgprot) = 0UL;
  pgprot = protection_map[VM_READ];
  
  if ((error = remap_page_range ((u_long)dst, src, PAGE_SIZE, pgprot)))
    return error;
  
  return (OK);
}

#endif

#if (0) /* How to handle this? */
/*
 * iti_init_fasync
 */

int
iti_init_fasync (itifg_t *iti,
		 iti_file_t *file,
		 iti_inode_t *inode,
		 int desc,
		 int mode)
{
  int error;

#if LINUX_VERSION_CODE < 0x020100
  error = fasync_helper (inode, file, mode, &iti->osp.cam_a_queue);
#elif LINUX_VERSION_CODE >= 0x020200
  error = fasync_helper (desc, file, mode, &iti->osp.cam_a_queue);
#endif
  
  if (mode)
    {
      if (!iti->osp.cam_a_queue)
	return -EINVAL;
    }
  else
    if (iti->osp.cam_a_queue)
      iti->osp.cam_a_queue = NULL;
  
  return OK;
}

/*
 * iti_kill_fasync
 */

int
iti_kill_fasync (itifg_t *iti, iti_sleep_t mode)
{

  switch (mode)
    {
    case ITI_CAMERA:
      if (!iti->osp.cam_a_queue)
	return ERROR;
#if LINUX_VERSION_CODE < 0x020300
      kill_fasync (iti->osp.cam_a_queue, SIGIO);
#elif LINUX_VERSION_CODE >= 0x020400
      kill_fasync (iti->osp.cam_a_queue, SIGIO, 0);
#endif
      break;
    case ITI_IOPORT:
      return -ENXIO;	/* not yet */
      break;
    default:
      return -EINVAL;
    }

  return OK;
}
#endif

/*
 * iti_os_interrupt
 */

int
iti_os_interrupt (void *data)
{
  iti_interrupt ((itifg_t*)data);

  return OK;
}

/*
 * iti_os_cam_timeout
 */

void
iti_os_cam_timeout (void *data)
{
  iti_cam_timeout ((itifg_t*)data);
}

/*
 * iti_os_iop_timeout
 */

void
iti_os_iop_timeout (void *data)
{
  iti_iop_timeout ((itifg_t*)data);
}

#if (0)
/*
 * iti_os_task_queue
 */

void
iti_os_task_queue (void *data)
{
  iti_task_queue ((int)((long)data));
}
#endif

/*
 * iti_os_open
 */

int
iti_os_open (dev_t dev, int oflags, int devtype, struct proc *p)
{
  return iti_open (NULL, dev); /* FIXME */
}

/*
 * iti_os_close
 */

int
iti_os_close (dev_t dev, int fflag, int devtype, struct proc *p)
{
  return iti_close (NULL, NULL, dev); /* FIXME */
}

/*
 * iti_os_read
 */

int
iti_os_read (dev_t dev, struct uio *uio, int ioflag)
{
  return iti_read (NULL, dev, uio->uio_iov[0].iov_base, /* FIXME */
		   uio->uio_iov[0].iov_len);
}

/*
 * iti_os_write
 */

int
iti_os_write (dev_t dev, struct uio *uio, int ioflag)
{
  return iti_write (NULL, dev, uio->uio_iov[0].iov_base, /* FIXME */
		    uio->uio_iov[0].iov_len);
}

#if (0) /* FIXME */
/*
 * iti_os_llseek
 */

static int
iti_os_lseek (struct inode *inode,
	      struct file *file,
	      off_t offset,
	      int mode)
{
  return iti_lseek (file, inode, offset, mode);
}
#endif

/* 
 * iti_os_ioctl
 */

int
iti_os_ioctl (dev_t dev, u_long cmd, caddr_t data,
	      int fflag, struct proc *p)
{
  int error, count, board, funct;
  
  itifg_t *iti;

  union iti_inparam_t *inparam = NULL;
  union iti_outparam_t *outparam = NULL;
  
  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error;
  iti = itifg[board];

#define IOCNUM(x)	(x & 0xff)	/* not defined in ioccom.h */

  /* validate module specific commands */
  if (IOCNUM(cmd) > 31 && IOCNUM(cmd) < 224)
    switch (MODULE_ID)
      {
      case ICP_AMFA:
	if (IOCNUM(cmd) > 63 || IOCGROUP(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMC1:
	if (IOCNUM(cmd) < 64 || IOCNUM(cmd) > 95 || IOCGROUP(cmd) != 'G')
	  return -EINVAL;
      break;
      case ICP_AMVS:
	if (IOCNUM(cmd) < 96 || IOCNUM(cmd) > 127 || IOCGROUP(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMDG:
	if (IOCNUM(cmd) < 128 || IOCNUM(cmd) > 159 || IOCGROUP(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMPV:
	if (IOCNUM(cmd) < 96 || IOCNUM(cmd) > 127 || IOCGROUP(cmd) != 'H')
	  return -EINVAL;
	break;
      case ITI_VMPCD:
	if (IOCNUM(cmd) < 128 || IOCNUM(cmd) > 159 || IOCGROUP(cmd) != 'H')
	  return -EINVAL;
	break;
      default:
	return -ENXIO;
      }

#define IOCDIR(x)	(x & IOC_DIRMASK)	/* not defined in ioccom.h */

  if (data)
    {
      /* structure allocation */
      if (IOCDIR(cmd) == IOC_OUT ||
	  IOCDIR(cmd) == IOC_INOUT)
	{
	  outparam = iti_malloc (sizeof(union iti_outparam_t), ITI_PAGED);
	  if (outparam == NULL)
	    return -EAGAIN;
	  ITI_PRINT3A("    outparam: 0x%p\n", outparam);
	}

      if (IOCDIR(cmd) == IOC_IN ||
	  IOCDIR(cmd) == IOC_INOUT)
	{
	  inparam = iti_malloc (sizeof(union iti_inparam_t), ITI_PAGED);
	  if (inparam == NULL)
	    return -EAGAIN;
	  ITI_PRINT3A("    inparam: 0x%p\n", inparam);
	}

      /* copyin */
      if (IOCDIR(cmd) == IOC_IN ||
	  IOCDIR(cmd) == IOC_INOUT)
	{
	  if (cmd == GIOC_DG_SET_LUT ||
	      cmd == HIOC_PCD_SET_LUT0 || cmd == HIOC_PCD_SET_LUT1)
	    {
	      /* data has to be copied in single-page chunks */
	      for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
		   count++)
		if ((error = iti_copyin ((void*)((caddr_t*)inparam +
						 count * PAGE_SIZE), 
					 (void*)(data + count * PAGE_SIZE),
					 PAGE_SIZE)))
		  return error;
	      /* if the structure is not a multiple of PAGE_SIZE */
	      if (sizeof(union dg_lut16_t) % PAGE_SIZE)
		if ((error = iti_copyin ((void*)((caddr_t)inparam +
						 count * PAGE_SIZE), 
					 (void*)(data + count * PAGE_SIZE),
					 sizeof(union dg_lut16_t)%PAGE_SIZE)))
		  return error;
	    }
	  else
	    {
	      if ((error = iti_copyin (inparam, (void*)data,
				       IOCPARM_LEN(cmd))))
		return error;
	    }
	}
    }
  else
    if (IOCDIR(cmd) == IOC_VOID)
      inparam = (void*)-1, outparam = (void*)-1;

  if ((error = iti_ioctl (NULL, dev, cmd, inparam, outparam))) /* FIXME */
    {
      if (outparam != (void*)-1)
	{
	  ITI_PRINT3A("    outparam: 0x%p\n", outparam);
	  iti_free (outparam, ITI_PAGED);
	}
      if (inparam != (void*)-1)
	{
	  ITI_PRINT3A("    inparam: 0x%p\n", inparam);
	  iti_free (inparam, ITI_PAGED);
	}
      return error;
    }

  if (data)
    {
      /* copyout */
      if (IOCDIR(cmd) == IOC_OUT ||
	  IOCDIR(cmd) == IOC_INOUT)
	{
	  if (cmd == GIOC_DG_GET_LUT ||
	      cmd == HIOC_PCD_GET_LUT0 || cmd == HIOC_PCD_GET_LUT1)
	    {
	      /* data has to be copied in single-page chunks */
	      for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
		   count++)
		if ((error = iti_copyout ((void*)((caddr_t)outparam +
						  count * PAGE_SIZE), 
					  (void*)(data + count * PAGE_SIZE),
					  PAGE_SIZE)))
		  return error;
	      /* if the structure is not a multiple of PAGE_SIZE */
	      if (sizeof(union dg_lut16_t) % PAGE_SIZE)
		if ((error = iti_copyout ((void*)((caddr_t)outparam +
						  count * PAGE_SIZE), 
					  (void*)(data + count * PAGE_SIZE),
					  sizeof(union dg_lut16_t)%PAGE_SIZE)))
		  return error;
	    }
	  else
	    {
	      if ((error = iti_copyout ((void*)data, outparam,
					IOCPARM_LEN(cmd))))
		return error;
	    }
	}

      /* free structures */
      if (outparam)
	{
	  ITI_PRINT3A("    outparam: 0x%p\n", outparam);
	  iti_free (outparam, ITI_PAGED);
	}
      if (inparam)
	{
	  ITI_PRINT3A("    inparam: 0x%p\n", inparam);
	  iti_free (inparam, ITI_PAGED);
	}
    }

  return OK;
}

/*
 * iti_os_poll
 */

int
iti_os_poll (dev_t dev, int events, struct proc *p)
{

  int error, result, mask, board, funct;

  itifg_t *iti;

  int type;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error;
  iti = itifg[board];

  type = 0;
  if (events & (POLLIN | POLLRDNORM))
    type |= ITI_SEL_IN;
  if (events & POLLPRI)
    type |= ITI_SEL_EX;

  mask = 0;
  result = iti_select (NULL, dev, type); /* FIXME */

  if (type & ITI_SEL_IN)
    {
      if (result & ITI_SEL_IN)
	{
	  iti->osp.cam_s_wait = FALSE;
	  mask |= (POLLIN | POLLRDNORM);
	}
      else
	iti->osp.cam_s_wait = TRUE;
    }

  if (type & ITI_SEL_EX)
    if (result & ITI_SEL_EX)
      mask |= POLLPRI;

  if (!mask)
    selrecord(p, &iti->osp.cam_s_queue);
  
  return mask;
}

/*
 * iti_os_mmap
 */

paddr_t
iti_os_mmap (dev_t dev, off_t offset, int nprot)
{
  int error;

  caddr_t base;
  size_t size;
  
  base = NULL;
  size = 0;
  
  if ((error = iti_mmap (NULL, dev, base, size))) /* FIXME */
    return error;
  
  return OK;
}

#if (0)
/*
 * iti_os_fasync
 */

static int
iti_os_fasync (int desc, struct file *file, int mode)
{
  return iti_fasync (NULL, NULL, desc, mode);
}

/*
 * iti_os_read_proc
 */

static int
iti_os_read_proc (char *buffer, char **start,
		  off_t offset, int length, int *eof, void *unused)
{
  int error, count;
  
  if ((error = iti_os_proc_info (buffer, &count)))
    return error;
  
  return count;
}

/*
 * iti_os_proc_info
 */

int
iti_os_proc_info (char *buffer, int *ret_count)
{
  int count, board , module;
  char mod_name[16];
  
  itifg_t *iti;

  board = 0;
  ITI_PRINT1("> iti_info\n");

  if (!buffer)
    return ERROR;

  count = 0;

  strcpy (mod_name, "Unknown");
  count += sprintf (&buffer[count], "%s %d.%d.%d-%d\n", PFS_VERSION_IDENT,
		     (iti_version_code & 0xFF000000) >> 24,
		     (iti_version_code & 0x00FF0000) >> 16,
		     (iti_version_code & 0x0000FF00) >> 8,
		     (iti_version_code & 0x000000FF)); 
  count += sprintf (&buffer[count], "%s %s\n", PFS_DATE_IDENT, __DATE__); 
  count += sprintf (&buffer[count], "%s %d\n", PFS_BOARDS_IDENT, boards);
  count += sprintf (&buffer[count], "%s", PFS_MODULES_IDENT);
  for (board = 0; board < boards; board++)
    {
      iti = itifg[board];

      for (module = 0;
	   module < sizeof(mod_names) / sizeof(mod_names[0]);
	   module++)
	if (MODULE_ID == mod_names[module].id)
	  strcpy (mod_name, mod_names[module].name);
      count += sprintf (&buffer[count], " %s", mod_name);
    }
  count += sprintf (&buffer[count], "\n");

  if (ret_count)
    *ret_count = count;

  ITI_PRINT1("< iti_info\n");
  return OK;
}
#endif

/*
 * iti_os_init
 */

int
iti_os_init (itifg_t *iti)
{
  ITI_PRINT1("> iti_os_init\n");

  /* FIXME: Nothing to do? */

  ITI_PRINT1("< iti_os_init\n");  
  return OK;
}

/*
 * the probe routine.
 */

int
iti_os_probe (struct device *parent, struct cfdata *match, void *aux)
{
  struct pci_attach_args *pa = aux;
  
  if (PCI_VENDOR(pa->pa_id) == ITI_PCI_VENDOR_ID_IMAGINGTECH &&
      (PCI_PRODUCT(pa->pa_id) == ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI ||
       PCI_PRODUCT(pa->pa_id) == ITI_PCI_DEVICE_ID_IMAGINGTECH_ASYNC ||
       PCI_PRODUCT(pa->pa_id) == ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG))
    return 1;
  
  return 0;
}

/*
 * iti_os_pci_init
 */

int
iti_os_pci_init (itifg_t *iti, void *data)
{
  u_long bhlcr, command;
  u_char latency;

  struct pci_attach_args *pa = (struct pci_attach_args*)data;
  
  ITI_PRINT1("> iti_os_pci_init\n");
  
  /* read device id */
  iti->device_id = pa->pa_device;
  
  /* read status register */
  iti->status =      
    pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);
  
  /* read revision id */
  iti->revision = 
    PCI_REVISION(pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_CLASS_REG));
  
  /* base addresses of various regions */
  /* FIXME: this is done in a different way */
  iti_memset (iti->address, 0, sizeof(iti->address));
  
  /* interrupt line */
  iti->irq_line = pa->pa_intrline;
  
  /*
   * PCI latency timer.  32 is a good value for 4 bus mastering slots, if
   * you have more than four, then 16 would probably be a better value.
   */

#define ITI_LATENCY_DEF	32
  
  /* latency timer (if too low -> correct) */
  bhlcr = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_BHLC_REG);
  latency = PCI_LATTIMER(bhlcr);
  
  if (latency < ITI_LATENCY_DEF)
    {
      latency = ITI_LATENCY_DEF;
      bhlcr &= ~((u_long)PCI_LATTIMER_MASK << PCI_LATTIMER_SHIFT);
      bhlcr |= (u_long)latency << PCI_LATTIMER_SHIFT;
      pci_conf_write(pa->pa_pc, pa->pa_tag, PCI_BHLC_REG, bhlcr);
    }
  
  /* enable bus master mode */
  command = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG);
  pci_conf_write(pa->pa_pc, pa->pa_tag, PCI_COMMAND_STATUS_REG,
		 command | PCI_COMMAND_MASTER_ENABLE);
  
  ITI_PRINT1A("    Device    : %d\n", pa->pa_device);
  ITI_PRINT1A("    Function  : %d\n", pa->pa_function);

  ITI_PRINT1("< iti_os_pci_init\n");
  return OK;
}

/*
 * iti_os_map_init
 */

int
iti_os_map_init (itifg_t *iti, void *data,
		 u_long *align_div, long *align_mod)
{
  int error;
  struct pci_attach_args *pa = (struct pci_attach_args*)data;

  ITI_PRINT1("> iti_os_map_init\n");

#define ITI_PCI_BADR0	PCI_MAPREG_START
#define ITI_PCI_BADR1	(PCI_MAPREG_START + 0x4)
#define ITI_PCI_BADR2	(PCI_MAPREG_START + 0x8)
#define ITI_PCI_BADR3	(PCI_MAPREG_START + 0xc)

  /* remap all register spaces to kernel memory */
  error = pci_mapreg_map(pa, ITI_PCI_BADR0,
			  (iti->regs_io ?
			   PCI_MAPREG_TYPE_MEM :
			   PCI_MAPREG_TYPE_IO)
			 | PCI_MAPREG_MEM_TYPE_32BIT, 0,
			 &iti->osp.icrt, &iti->osp.icrh,
			 (u_long*)&iti->v_icr_base, NULL);
  if (error)
    {
      iti_printw ("unable to ioremap interface control regs.\n");
      return -EFAULT;
    }

  switch (iti->device_id)
    {
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI:
      error = pci_mapreg_map(pa, ITI_PCI_BADR1,
			      (iti->regs_io ?
			       PCI_MAPREG_TYPE_MEM :
			       PCI_MAPREG_TYPE_IO)
			      | PCI_MAPREG_MEM_TYPE_32BIT, 0,
			      &iti->osp.crt, &iti->osp.crh,
			      (u_long*)&iti->v_cr_base, NULL);
      if (error)
	{
	  iti_printw ("unable to ioremap control regs.\n");
	  return -EFAULT;
	}
      error = pci_mapreg_map(pa, ITI_PCI_BADR2,
			      PCI_MAPREG_TYPE_MEM
			      | PCI_MAPREG_MEM_TYPE_32BIT, 0,
			      &iti->osp.modt, &iti->osp.modh,
			      (u_long*)&iti->v_mod_base, NULL);
      if (error)
	{
	  iti_printw ("unable to ioremap acquisition modul regs.\n");
	  return -EFAULT;
	}
      error = pci_mapreg_map(pa, ITI_PCI_BADR3,
			      PCI_MAPREG_TYPE_MEM
			      | PCI_MAPREG_MEM_TYPE_32BIT, 0,
			      &iti->osp.memt, &iti->osp.memh,
			      (u_long*)&iti->v_mem_base,
			      &iti->osp.mems);
      if (error)
	{
	  iti_printw ("unable to ioremap main board memory.\n");
	  return -EFAULT;
	}
      break;

    case ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG:
      error = pci_mapreg_map(pa, ITI_PCI_BADR1,
			      (iti->regs_io ?
			       PCI_MAPREG_TYPE_MEM :
			       PCI_MAPREG_TYPE_IO)
			      | PCI_MAPREG_MEM_TYPE_32BIT, 0,
			      &iti->osp.crt, &iti->osp.crh,
			      (u_long*)&iti->v_cr_base, NULL);
      if (error)
	{
	  iti_printw ("unable to ioremap control regs.\n");
	  return -EFAULT;
	}
      /* not used at the PCDig */
      iti->v_mod_base = NULL;
      
      error = pci_mapreg_map(pa, ITI_PCI_BADR3,
			      PCI_MAPREG_TYPE_MEM
			      | PCI_MAPREG_MEM_TYPE_32BIT, 0,
			      &iti->osp.memt, &iti->osp.memh,
			      (u_long*)&iti->v_mem_base,
			      &iti->osp.mems);
      if (error)
	{
	  iti_printw ("unable to ioremap main board memory.\n");
	  return -EFAULT;
	}
      break;

    default:
      return -ENXIO;
    }

  ITI_PRINT1("< iti_os_map_init\n");
  return OK;
}

/*
 * iti_os_attach
 */

static void
iti_os_attach (struct device *parent, struct device *self, void *aux)
{
  int error;

  pci_intr_handle_t ih;
  const char *intrstr;

  itifg_t *iti;

  struct pci_attach_args *pa = aux;
  
  ITI_PRINT1("> iti_os_attach\n");

  /* check PCI - configuration space for possible devices */

  iti_printia ("configure device %d.\n", boards);
  boards++;
  
  /* init per board driver */
  if ((error = iti_attach (boards, aux)))
    return;
  
  iti = itifg[boards];
  
  /* register per board interrupts */
  if (pci_intr_map(pa->pa_pc, pa->pa_intrtag, pa->pa_intrpin,
		   pa->pa_intrline, &ih))
    {
      iti_printwa ("'map intr (intr. %d)' failed.\n",
		   iti->irq_line);
      return;
    }
    
  intrstr = pci_intr_string(pa->pa_pc, ih);
  iti->osp.ih = pci_intr_establish(pa->pa_pc, ih, ITIFGIPL,
				   iti_os_interrupt, &iti);
  if (iti->osp.ih == NULL)
    {
      iti_printwa ("'establish intr (intr. %d)' failed.\n",
		   iti->irq_line);
      return;
    }
  
  /* print version */
  iti_printia ("version %d.%d.%d-%d.\n",
	       (iti_version_code & 0xFF000000) >> 24,
	       (iti_version_code & 0x00FF0000) >> 16,
	       (iti_version_code & 0x0000FF00) >> 8,
	       (iti_version_code & 0x000000FF));
  iti_printia ("%s (linux %s).\n", __DATE__, OS_RELEASE);
  
  ITI_PRINT1("< iti_os_attach\n");  
}

/* these functions are only called, if NOT using the inline equivalents */

u_short
iti_readw (u_short *addr) 
{
  bus_space_read_2 (
}
void
iti_writew (u_short value, u_short *addr)
{
  writew (value, addr);
}
u_long
iti_readl (u_long *addr) 
{
  return readl (addr);
}
void
iti_writel (u_long value, u_long *addr)
{
  writel (value, addr);
}
u_short
iti_inw (u_short port) 
{
  return inw (port);
}
void
iti_outw (u_short value, u_short port)
{
  outw (value, port);
}
u_long
iti_inl (u_short port) 
{
  return inl (port);
}
void
iti_outl (u_long value, u_short port)
{
  outl (value, port);
}

#if (0)
void*
iti_memset_io (void *dst, int val, size_t size)
{
  return memset_io (dst, val, size);
}
#endif

void
iti_udelay (u_long time)
{
  delay (time);
}

int
iti_copyout (void *dst, void *src, size_t size)
{
  return copyout (src, dst, size))
}
int
iti_copyin (void *dst, const void *src, size_t size)
{
  return copyin (src, dst, size))
}

void*
iti_memcpy (void* dst, const void* src, size_t size)
{
  return bcopy (src, dst, size);
}
void*
iti_memset (void* dst, int value, size_t size)
{
  if (value)
    return memset (dst, value, size);
  else
    bzero (dst, size);
}
char*
iti_strcpy (char* dst, const char* src)
{
  return strcpy (dst, src);
}
int
iti_strcmp (const char* dst, const char* src)
{
  return strcmp (dst, src);
}

void
iti_cleanup (void)
{
  cleanup_module ();
}
#endif
